__version__ = "0.9.0"

from . import lib
from .lib import (
    EncodedSequence,
    StripedSequence,
    CountMatrix,
    WeightMatrix,
    ScoringMatrix,
    Motif,
    TransfacMotif,
    UniprobeMotif,
    JasparMotif,
    StripedScores,
    Scanner,
    Hit,
    Loader,
    create,
    stripe,
    scan,
    load
)

__author__ = lib.__author__
__license__ = "MIT OR GPL-3.0-or-later"
__doc__ = lib.__doc__



import unittest
import io
import os
import textwrap
import tempfile
import pathlib

import lightmotif


class _TestLoad():
    
    def test_load_file(self):
        text = textwrap.dedent(self.text).encode()
        buffer = io.BytesIO(text)
        motifs = list(lightmotif.load(buffer, self.format))
        self.assertEqual(len(motifs), self.length)
        self.assertEqual(motifs[0].name, self.first)

    def test_load_filename(self):
        text = textwrap.dedent(self.text).encode()
        with tempfile.NamedTemporaryFile("r+b") as f:
            f.write(text)
            f.flush()
            motifs = list(lightmotif.load(f.name, self.format))
        self.assertEqual(len(motifs), self.length)
        self.assertEqual(motifs[0].name, self.first)

    def test_load_filename_bytes(self):
        text = textwrap.dedent(self.text).encode()
        with tempfile.NamedTemporaryFile("r+b") as f:
            f.write(text)
            f.flush()
            motifs = list(lightmotif.load(os.fsencode(f.name), self.format))
        self.assertEqual(len(motifs), self.length)
        self.assertEqual(motifs[0].name, self.first)

    def test_load_path(self):
        text = textwrap.dedent(self.text).encode()
        with tempfile.NamedTemporaryFile("r+b") as f:
            f.write(text)
            f.flush()
            path = pathlib.Path(f.name)
            motifs = list(lightmotif.load(path, self.format))
        self.assertEqual(len(motifs), self.length)
        self.assertEqual(motifs[0].name, self.first)


class TestJASPAR(_TestLoad, unittest.TestCase):
    format = "jaspar"
    length = 1
    first = "MA0001.3"
    text = """
    >MA0001.3	AGL3
    0      0     82     40     56     35     65     25     64      0
    92    79      1      4      0      0      1      4      0      0
    0      0      2      3      1      0      4      3     28     92
    3     16     10     48     38     60     25     63      3      3
    """

class TestJASPAR16(_TestLoad, unittest.TestCase):
    format = "jaspar16"
    length = 2
    first = "MA0001.3"
    text = """
    >MA0001.3	AGL3
    A  [     0      0     82     40     56     35     65     25     64      0 ]
    C  [    92     79      1      4      0      0      1      4      0      0 ]
    G  [     0      0      2      3      1      0      4      3     28     92 ]
    T  [     3     16     10     48     38     60     25     63      3      3 ]
    >MA0017.3	NR2F1
    A  [  7266   6333   8496      0      0      0      0  12059   5116   3229   3276   5681 ]
    C  [  1692    387      0     30      0      0  12059      0   3055   2966   2470   2912 ]
    G  [  1153   4869   3791  12059  12059      0      0     91   1618   4395   3886  36863 ]
    T  [  1948    469      0      0      0  12059      0      0   2270   1469   2427   3466 ]
    """

class TestTRANSFAC(_TestLoad, unittest.TestCase):
    format = "transfac"
    length = 1
    first = "M00005"
    text = """
    NA  M00005
    P0      A      C      G      T
    01      3      0      0      2      W
    02      1      1      3      0      G
    03      3      1      1      0      A
    04      2      1      2      0      R
    05      1      2      0      2      Y
    06      0      5      0      0      C
    07      5      0      0      0      A
    08      0      0      5      0      G
    09      0      5      0      0      C
    10      0      0      1      4      T
    11      0      1      4      0      G
    12      0      2      1      2      Y
    13      1      0      3      1      G
    14      0      0      5      0      G
    15      1      1      1      2      N
    16      1      4      0      0      C
    17      2      1      1      1      N
    18      0      0      3      2      K
    //
    """.lstrip("\n")


from . import (
    test_doctest, 
    test_pipeline,
    test_scanner, 
    test_sequence, 
    test_pvalue,
    test_load
)


def load_tests(loader, suite, pattern):
    suite.addTests(loader.loadTestsFromModule(test_doctest))
    suite.addTests(loader.loadTestsFromModule(test_pipeline))
    suite.addTests(loader.loadTestsFromModule(test_pvalue))
    suite.addTests(loader.loadTestsFromModule(test_scanner))
    suite.addTests(loader.loadTestsFromModule(test_sequence))
    suite.addTests(loader.loadTestsFromModule(test_load))
    return suite



import os
import time
import typing
import statistics
import sys

import Bio.motifs
import Bio.Seq
import numpy
import pymemesuite.common
import pymemesuite.fimo
import MOODS.scan
from fs.filesize import binary

sys.path.append(os.path.realpath(os.path.join(__file__, "..", "..", "..")))
import lightmotif


class Timer(typing.ContextManager["Timer"]):
    def __init__(self):
        self.start: Optional[float] = None
        self.end: Optional[float] = None

    def __enter__(self) -> "Timer":
        self.end = None
        self.start = time.time()
        return self

    def __exit__(self, exc_value, exc_ty, tb) -> bool:
        self.end = time.time()
        return False

    def total(self) -> float:
        if self.end is None:
            raise RuntimeError("Timer has not stopped")
        return self.end - self.start


N = 10
instances = ["GTTGACCTTATCAAC", "GTTGATCCAGTCAAC"]

with open("/home/althonos/Code/lightmotif/lightmotif/benches/ecoli.txt") as f:
    seq = f.read()


# --- lightmotif ---------------------------------------------------------------

eseq = lightmotif.EncodedSequence(seq)
sseq = eseq.stripe()

motif = lightmotif.create(instances)
frequencies = motif.counts.normalize(0.1)
pssm = frequencies.log_odds()

times = []
for _ in range(N):
    with Timer() as timer:
        scores = pssm.calculate(sseq)
    times.append(timer.total() * 1e9)

avg = statistics.mean(times)
dev = max(times) - min(times)
speed = int(len(seq) * 1e9 / avg)
print(
    f"lightmotif (avx2):".ljust(20),
    f"{avg:,.0f} ns/iter".rjust(20),
    f"(+/- {dev:,.0f})".rjust(18),
    "=",
    f"{binary(speed)}/s".rjust(11),
)
# print(numpy.asarray(scores).ravel()[: len(scores)])
# print(numpy.max(scores))

# --- Bio.motifs ---------------------------------------------------------------

motif = Bio.motifs.create(instances)
frequencies = motif.counts.normalize(0.1)
pssm = frequencies.log_odds()

times = []
for _ in range(N):
    with Timer() as timer:
        scores = pssm.calculate(seq)
    times.append(timer.total() * 1e9)

avg = statistics.mean(times)
dev = max(times) - min(times)
speed = int(len(seq) * 1e9 / avg)
print(
    f"Bio.motifs:".ljust(20),
    f"{avg:,.0f} ns/iter".rjust(20),
    f"(+/- {dev:,.0f})".rjust(18),
    "=",
    f"{binary(speed)}/s".rjust(11),
)
# print(numpy.asarray(scores))
# print(numpy.max(scores))


# --- MOODS.scan ---------------------------------------------------------------

motif = Bio.motifs.create(instances)
pssm = motif.counts.normalize(0.1).log_odds()

# create MOODS scanner
scanner = MOODS.scan.Scanner(7)
m = [[pssm[x] for x in "ATGC"]]
scanner.set_motifs(m, (0.25, 0.25, 0.25, 0.25), [0])

times = []
for _ in range(N):
    with Timer() as timer:
        scanner.scan(seq)
    times.append(timer.total() * 1e9)

avg = statistics.mean(times)
dev = max(times) - min(times)
speed = int(len(seq) * 1e9 / avg)
print(
    f"MOODS.scan:".ljust(20),
    f"{avg:,.0f} ns/iter".rjust(20),
    f"(+/- {dev:,.0f})".rjust(18),
    "=",
    f"{binary(speed)}/s".rjust(11),
)

# --- PyMEMEsuite --------------------------------------------------------------

alphabet = pymemesuite.common.Alphabet.dna()
background = pymemesuite.common.Background.from_uniform(alphabet)

counts = pymemesuite.common.Matrix.zeros(len(instances[0]), len(alphabet.symbols))
for instance in instances:
    for i, base in enumerate(instance):
        j = alphabet.symbols.index(base)
        counts[i, j] += 1

frequencies = pymemesuite.common.Matrix.zeros(len(instances[0]), len(alphabet.symbols))
for i in range(len(instances[0])):
    n = counts[i].sum()
    for j in range(len(alphabet.symbols)):
        frequencies[i, j] = counts[i, j] / n

motif = pymemesuite.common.Motif(alphabet, frequencies=frequencies)
pssm = motif.build_pssm(background)
fimo = pymemesuite.fimo.FIMO(both_strands=False)
mmsq = pymemesuite.common.Sequence(seq)

times = []
for _ in range(N):
    with Timer() as timer:
        fimo.score_pssm(pssm, [mmsq])
    times.append(timer.total() * 1e9)

avg = statistics.mean(times)
dev = max(times) - min(times)
speed = int(len(seq) * 1e9 / avg)
print(
    f"pymemesuite.fimo:".ljust(20),
    f"{avg:,.0f} ns/iter".rjust(20),
    f"(+/- {dev:,.0f})".rjust(18),
    "=",
    f"{binary(speed)}/s".rjust(11),
)



import typing
from typing import Dict, List, Iterable, Union, Optional, BinaryIO, Iterator, Generic
from os import PathLike

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore

__version__: str
__author__: str

AVX2_SUPPORTED: bool

FORMAT = Literal["jaspar", "jaspar16", "uniprobe", "transfac"]

class EncodedSequence:
    def __init__(self, sequence: str, *, protein: bool = False) -> None: ...
    def __str__(self) -> str: ...
    def __len__(self) -> int: ...
    def __copy__(self) -> EncodedSequence: ...
    def __buffer__(self, flags: int) -> memoryview: ...
    @property
    def protein(self) -> bool: ...
    def copy(self) -> EncodedSequence: ...
    def stripe(self) -> StripedSequence: ...

class StripedSequence:
    def __copy__(self) -> StripedSequence: ...
    def __buffer__(self, flags: int) -> memoryview: ...
    @property
    def protein(self) -> bool: ...
    def copy(self) -> StripedSequence: ...

class CountMatrix:
    def __init__(
        self, values: Dict[str, Iterable[int]], *, protein: bool = False
    ) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __len__(self) -> int: ...
    def __getitem__(self, index: int) -> List[int]: ...
    @property
    def protein(self) -> bool: ...
    def normalize(
        self, pseudocount: Union[None, float, Dict[str, float]] = None
    ) -> WeightMatrix: ...

class WeightMatrix:
    def __eq__(self, other: object) -> bool: ...
    def __len__(self) -> int: ...
    def __getitem__(self, index: int) -> List[int]: ...
    @property
    def protein(self) -> bool: ...
    def log_odds(
        self, background: Union[None, Dict[str, float]] = None, base: float = 2.0
    ) -> ScoringMatrix: ...

class ScoringMatrix:
    def __init__(
        self,
        values: Dict[str, Iterable[float]],
        background: Optional[Dict[str, float]] = None,
        *,
        protein: bool = False,
    ) -> None: ...
    def __len__(self) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def __buffer__(self, flags: int) -> memoryview: ...
    @property
    def protein(self) -> bool: ...
    def calculate(self, sequence: StripedSequence) -> StripedScores: ...
    def pvalue(self, score: float) -> float: ...
    def score(self, pvalue: float) -> float: ...
    def reverse_complement(self) -> ScoringMatrix: ...

class StripedScores:
    def __len__(self) -> int: ...
    def __getitem__(self, index: int) -> float: ...
    def __buffer__(self, flags: int) -> memoryview: ...
    def threshold(self, threshold: float) -> List[int]: ...
    def max(self) -> Optional[float]: ...
    def argmax(self) -> Optional[int]: ...

class Motif:
    @property
    def counts(self) -> Optional[CountMatrix]: ...
    @property
    def pwm(self) -> WeightMatrix: ...
    @property
    def pssm(self) -> ScoringMatrix: ...
    @property
    def name(self) -> Optional[str]: ...
    @property
    def protein(self) -> bool: ...

class TransfacMotif(Motif):
    @property
    def counts(self) -> CountMatrix: ...

class JasparMotif(Motif):
    @property
    def counts(self) -> CountMatrix: ...
    @property
    def name(self) -> str: ...
    @property
    def description(self) -> Optional[str]: ...

class UniprobeMotif(Motif):
    @property
    def counts(self) -> None: ...

class Scanner(Iterator[Hit]):
    def __init__(
        self,
        pssm: ScoringMatrix,
        sequence: StripedSequence,
        threshold: float = 0.0,
        block_size: int = 256,
    ) -> None: ...
    def __iter__(self) -> Scanner: ...
    def __next__(self) -> Hit: ...

class Hit:
    @property
    def position(self) -> int: ...
    @property
    def score(self) -> float: ...

M = typing.TypeVar("M", bound=Motif)

class Loader(Generic[M], Iterator[M]):
    def __init__(
        self,
        file: Union[BinaryIO, PathLike[str]],
        format: Literal["jaspar16"],
        *,
        protein: bool = False,
    ) -> None: ...
    def __iter__(self) -> Loader[M]: ...
    def __next__(self) -> M: ...

def create(
    sequences: Iterable[str], *, protein: bool = False, name: Optional[str] = None
) -> Motif: ...
def stripe(sequence: str, *, protein: bool = False) -> StripedSequence: ...
def scan(
    pssm: ScoringMatrix,
    sequence: StripedSequence,
    *,
    threshold: float = 0.0,
    block_size: int = 256,
) -> Scanner: ...
@typing.overload
def load(
    file: Union[BinaryIO, PathLike[str]],
    format: Literal["jaspar"],
    *,
    protein: bool = False,
) -> Loader[JasparMotif]: ...
@typing.overload
def load(
    file: Union[BinaryIO, PathLike[str]],
    format: Literal["jaspar16"],
    *,
    protein: bool = False,
) -> Loader[JasparMotif]: ...
@typing.overload
def load(
    file: Union[BinaryIO, PathLike[str]],
    format: Literal["uniprobe"],
    *,
    protein: bool = False,
) -> Loader[UniprobeMotif]: ...
@typing.overload
def load(
    file: Union[BinaryIO, PathLike[str]],
    format: Literal["transfac"],
    *,
    protein: bool = False,
) -> Loader[TransfacMotif]: ...
@typing.overload
def load(
    file: Union[BinaryIO, PathLike[str]],
    format: FORMAT = "jaspar",
    *,
    protein: bool = False,
) -> Loader[Motif]: ...
